
#include "ext_protocol.h"
#include "ext_protocol_cmd.h"
#include "ext_protocol_crc.h"

static extCmdHandlerTab_t extCmdHandlerTab[EXT_PROTOCOL_CMD_MAX_NUM];        //the table of cmdid handle function
static extSendHandler_t ExtProtocolSend;
static uint8_t extSeqNum;



static uint16_t ExtProtocolPack(uint8_t *p_out, uint8_t *p_in, uint16_t len, uint16_t cmd_id)
{
    uint16_t headSize = EXT_PROTOCOL_HEADER_SIZE;
    uint16_t frameSize = len + EXT_HEADER_CRC_LEN + sizeof(cmd_id);

    memcpy(p_out + headSize, &cmd_id, sizeof(cmd_id));
    Ext_Append_CRC8_Check_Sum(p_out, headSize);
    memcpy(p_out + headSize + sizeof(cmd_id), p_in, len);
    Ext_Append_CRC16_Check_Sum(p_out, frameSize);

    return frameSize;
}


/**
*@func:     ExtProtocolTransmit
*@brief:    pc send protocol to main control
*@param:    cmd_id: the protocol cmd id
*           pdata: the point about data buffer
*           len: the length of the data buffer
*/
void ExtProtocolTransmit(uint16_t cmd_id, void* p_buf, uint16_t len)
{
    uint8_t txBuf[EXT_PROTOCOL_FRAME_MAX_SIZE]={0};
    extHeader_t *pHeader = (extHeader_t *)txBuf;
    uint16_t frameSize;

    pHeader->sof                    = EXT_PROTOCOL_HEADER;
    pHeader->length                 = len;
    pHeader->seq_num                = extSeqNum++;

    frameSize = ExtProtocolPack(txBuf, p_buf, len, cmd_id);
    ExtProtocolSend(txBuf, frameSize);
}


/**
*@func:     ExtProtocolCmdFuncTableInit
*@brief:    init the table of protocol cmdid handler function
*@param:    none
*@ret:      none
*/
void  ExtProtocolCmdHandleTableInit(void)
{
    for(int i = 0 ; i < EXT_PROTOCOL_CMD_MAX_NUM;i++)
    {
        extCmdHandlerTab[i].cmd_id = NULL;
        extCmdHandlerTab[i].handler = NULL;
    }
}

/**
*@func:     ExtProtocolCmdHandleRegister
*@brief:    register cmd handler function in cmd id handler table(gModuleCmdFunctionTable)
*@param:    none
*@note:     you must init the gModuleCmdFunctionTable through calling "ModuleCmdIdFuncTableInit()"
*           before registering your cmdid hander function
*@ret:      0: success
*           -1:failed
*/
int ExtProtocolCmdHandleRegister(uint16_t cmd_id, extCmdHandle_t fun_ptr)
{
    int i;
    for(i = 0;i < EXT_PROTOCOL_CMD_MAX_NUM; i++)
    {
        if((extCmdHandlerTab[i].cmd_id == NULL) && (extCmdHandlerTab[i].handler == NULL))
        {
            extCmdHandlerTab[i].cmd_id = cmd_id;
            extCmdHandlerTab[i].handler = fun_ptr;
            return 0;
        }
    }
    return -1;   // if  run here, it indicate the register fail.
}

void ExtProtocolSendDataHandleRegister(extSendHandler_t fun_ptr)
{
    ExtProtocolSend = fun_ptr;
}

/**
*@func:     ExtProtocolCmdHandle
*@brief:    cmdid handler function
*@param:    none
*@author:   damom.li
*/
static void ExtProtocolCmdHandle(uint8_t* p_buf, uint16_t len)
{
    uint16_t *p_cmdid = (uint16_t*)&p_buf[EXT_PROTOCOL_HEADER_SIZE];
    uint8_t i=0;

    for( i = 0; (extCmdHandlerTab[i].handler != NULL) && (i < EXT_PROTOCOL_CMD_MAX_NUM); i++ )
    {
        if( *p_cmdid == extCmdHandlerTab[i].cmd_id )
        {
            extCmdHandlerTab[i].handler(p_buf, len);
            //printf("protocol exe: 0x%X\r\n", p_cmdid);
        }
    }
}

void ExtProtocolUnpackObjInit(extUnpackObj_t *p_obj, FIFO_S_t * fifo)
{
    p_obj->p_fifo = fifo;
    p_obj->index = 0;
    p_obj->unpack_step = 0;
    p_obj->data_len = 0;
    p_obj->p_head = (extHeader_t *)p_obj->protocol_packet;
}


/*
*@func:ExtProtocolUnpack
*@brief:  unpack protocol package
*@param:
*@ret :1  crc8 or cr16 check failed
*      0  unpack successfed or receiving data is not finished.
*/
uint8_t  ExtProtocolUnpack(extUnpackObj_t *pObj)
{
    uint8_t byte = 0;
    uint8_t headerLen = EXT_PROTOCOL_HEADER_SIZE;
    uint8_t unpackResult = 0;
    while(FIFO_S_CountUsed(pObj->p_fifo))
    {
        byte = FIFO_S_Get(pObj->p_fifo);
        switch(pObj->unpack_step)
        {
            case 0:  // sof
                if(byte == EXT_PROTOCOL_HEADER)
                {
                    pObj->unpack_step = 1;
                    pObj->protocol_packet[pObj->index++] = byte;
                }
                else
                {
                    pObj->index = 0;
                }
                break;
            case 1:
                    pObj->data_len = byte;
                    pObj->protocol_packet[pObj->index++] = byte;
                    pObj->unpack_step = 2;
                break;
            case 2:    // data_len high
                    pObj->data_len |= (byte<<8);
                    pObj->protocol_packet[pObj->index++] = byte;

                    if(pObj->data_len < (EXT_PROTOCOL_FRAME_MAX_SIZE - EXT_HEADER_CRC_LEN))
                    {
                        pObj->unpack_step = 3;
                    }
                    else
                    {
                        pObj->unpack_step = 0;
                        pObj->index = 0;
                    }
                    break;

            case 3:
                    pObj->protocol_packet[pObj->index++] = byte;

                    if (pObj->index == headerLen)
                    {
                        if(Ext_Verify_CRC8_Check_Sum(pObj->protocol_packet, headerLen))
                        {
                            pObj->unpack_step = 4;
                        }
                        else
                        {
                            unpackResult = 1;
                            pObj->unpack_step = 0;
                            pObj->index = 0;
                        }
                    }
                    break;

            case 4:
                if(pObj->index < (pObj->data_len + EXT_HEADER_CRC_CMDID_LEN))
                {
                     pObj->protocol_packet[pObj->index++] = byte;
                }
                if(pObj->index >= (pObj->data_len + EXT_HEADER_CRC_CMDID_LEN))
                {
                    pObj->index = 0;
                    pObj->unpack_step = 0;

                    if(Ext_Verify_CRC16_Check_Sum(pObj->protocol_packet, pObj->data_len + EXT_HEADER_CRC_CMDID_LEN))
                    {
                        ExtProtocolCmdHandle(pObj->protocol_packet, pObj->data_len + EXT_HEADER_CRC_CMDID_LEN);
                    }
                    else
                    {
                        unpackResult = 1;
                    }
                }
                break;

            default:
                pObj->unpack_step = 0;
                pObj->index = 0;
                break;
        }
    }
    return unpackResult;
}



/************************ (C) COPYRIGHT 2017 DJI RoboMaster *****END OF FILE****/
