//******************************************************************************************
//!
//! \file   FIFO.c
//! \brief  Genernal FIFO Model Interface.
//!         You can use uniform FIFO Model to manager Any type of data element.
//! \author cedar
//! \changed damom.li
//! \changed sky.huang
//! \date   2016-11-03//! \date   2013-12-16
//! \email  xuesong5825718@gmail.com
//!
//! \license
//!
//! Copyright (c) 2013 Cedar MIT License
//!
//! Permission is hereby granted, free of charge, to any person obtaining a copy
//! of this software and associated documentation files (the "Software"), to deal
//! in the Software without restriction, including without limitation the rights to
//! use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of
//! the Software, and to permit persons to whom the Software is furnished to do so,
//! subject to the following conditions:
//!
//! The above copyright notice and this permission notice shall be included in all
//! copies or substantial portions of the Software.
//!
//! THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
//! IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
//! FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
//! AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
//! LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
//! OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS
//! IN THE SOFTWARE.
///
//******************************************************************************************
#include "fifo.h"

//******************************************************************************************
//!                     ASSERT MACRO
//******************************************************************************************
#ifndef ASSERT

#ifdef  NDEBUG
#define ASSERT(x)
#else
#define ASSERT(x) do {while(!(x));} while(0)
#endif

#endif  // ASSERT

//******************************************************************************************
//!                     Macro Function
//******************************************************************************************
#define MASTER_INT_ENABLE()            do{__enable_irq();}while(0)
#define MASTER_INT_DISABLE()           do{__disable_irq();}while(0)



#define CRITICAL_SETCION_ENTER()    MASTER_INT_DISABLE()
#define CRITICAL_SETCION_EXIT()     MASTER_INT_ENABLE()




#ifdef USE_DYNAMIC_MEMORY
//******************************************************************************************
//
//! \brief  Create An New FIFO Instance(in Single Mode).
//! This function allocate enought room for N blocks fifo elements, then return the pointer
//! of FIFO.
//!
//! \param  [in] uint_cnt is count of fifo elements.
//! \retval The Pointer of FIFO instance, return NULL is failure to allocate memory.
//!
//! \note   -# You must enable USE_MEMORY_ALLOC macro and ensure your system have <stdlib.h>
//!            Header file before use this function.
//! \note   -# Functions FIFO_Create and FIFO_Destory must be used in pairs.
//!
//******************************************************************************************
FIFO_S_t* FIFO_S_Create(uint32_t uint_cnt)
{
    FIFO_S_t* p_fifo        = NULL;         //!< FIFO Pointer
    uint8_t*  p_base_addr   = NULL;         //!< Memory Base Address
    //! Check input parameters.
    ASSERT(0 != uint_cnt);

    //! Allocate Memory for pointer of new FIFO Control Block.
    p_fifo = (FIFO_S_t*) pvPortMalloc(sizeof(FIFO_S_t));      /* malloc cannot compatible with __use_no_semihosting */
    p_base_addr = pvPortMalloc(uint_cnt);

    if ((NULL == p_fifo) || (NULL == p_base_addr))
    {
        //! Allocate Failure, exit now.
        return (NULL);
    }

    //! Initialize General FIFO Module.
    FIFO_S_Init(p_fifo, p_base_addr, uint_cnt);

    return (p_fifo);
}

//******************************************************************************************
//
//! \brief  Destory FIFO Instance(in Single Mode).
//!  This function release memory, then reinit the FIFO struct.
//!
//! \param  [in] p_fifo is the pointer of FIFO instance
//! \retval None.
//!
//! \note   -# You must enable USE_MEMORY_ALLOC macro and ensure your system have <stdlib.h>
//!            Header file before use this function.
//
//******************************************************************************************
void FIFO_S_Destory(FIFO_S_t* p_fifo)
{
    //! Check input parameters.
    ASSERT(NULL != p_fifo);
    ASSERT(NULL != p_fifo->p_start_addr);

    //! Free FIFO memory
    vPortFree(p_fifo->p_start_addr);
    //! Free FIFO Control Block memory.
    vPortFree(p_fifo);

    return;     //!< Success
}

#endif // USE_DYNAMIC_MEMORY


//******************************************************************************************
//
//! \brief  Initialize an static FIFO struct(in single mode).
//!
//! \param  [in] p_fifo is the pointer of valid FIFO instance.
//! \param  [in] p_base_addr is the base address of pre-allocate memory, such as array.
//! \param  [in] uint_cnt is count of fifo elements.
//! \retval 0 if initialize successfully, otherwise return -1.
//
//******************************************************************************************
int FIFO_S_Init(FIFO_S_t* p_fifo, void* p_base_addr, uint32_t uint_cnt)
{
    //! Check input parameters.
    ASSERT(NULL != p_fifo);
    ASSERT(NULL != p_base_addr);
    ASSERT(0    != uint_cnt);

    //! Initialize FIFO Control Block.
    p_fifo->p_start_addr    = (uint8_t*) p_base_addr;
    p_fifo->p_end_addr      = (uint8_t*) p_base_addr + uint_cnt - 1;
    p_fifo->buffer_size     = uint_cnt;
    p_fifo->free_num        = uint_cnt;
    p_fifo->used_num        = 0;
    p_fifo->read_index      = 0;
    p_fifo->write_index     = 0;

    return (0);
}



//******************************************************************************************
//
//! \brief  Put an element into FIFO(in single mode).
//!
//! \param  [in]  p_fifo is the pointer of valid FIFO.
//! \param  [in]  element is the data element you want to put
//!
//! \retval 0 if operate successfully, otherwise return -1.
//
//******************************************************************************************
int FIFO_S_Put(FIFO_S_t* p_fifo, uint8_t element)
{
    //! Check input parameters.
    ASSERT(NULL != p_fifo);

    // Full ?
    if (0 == p_fifo->free_num)
    {
        //! Error, FIFO is full!
        return (-1);
    }
    CRITICAL_SETCION_ENTER();
    p_fifo->p_start_addr[p_fifo->write_index] = element;
    p_fifo->free_num--;
    p_fifo->used_num++;
    p_fifo->write_index++;
    if(p_fifo->write_index == p_fifo->buffer_size) p_fifo->write_index = 0;
    CRITICAL_SETCION_EXIT();

    return (0);
}

//******************************************************************************************
//
//! \brief  Put some elements into FIFO(in single mode).
//!
//! \param  [in]  p_fifo is the pointer of valid FIFO.
//! \param  [in]  p_source is the data element you want to put
//! \param  [in]  the number of elements
//! \retval the number of really write data, otherwise return -1.
//
//******************************************************************************************
int FIFO_S_Puts(FIFO_S_t *p_fifo, uint8_t *p_source, uint32_t len)
{
    int      retval;
    uint32_t len_to_end;
    uint32_t len_from_start;

    ASSERT(NULL != p_fifo);

    if (NULL == p_source)
        return -1;

    if (0 == p_fifo->free_num)
        return 0;

    CRITICAL_SETCION_ENTER();
    len = (len < p_fifo->free_num) ? len : p_fifo->free_num;
    len_to_end = p_fifo->buffer_size - p_fifo->write_index;

    if(len_to_end >= len) //no rollback
    {
        len_to_end = len;
        memcpy(&p_fifo->p_start_addr[p_fifo->write_index], p_source, len_to_end);
        p_fifo->write_index = (p_fifo->write_index + len_to_end);
    }
    else  //rollback
    {
        len_from_start = len - len_to_end;
        memcpy(&p_fifo->p_start_addr[p_fifo->write_index], p_source, len_to_end);
        memcpy(&p_fifo->p_start_addr[0], p_source + len_to_end, len_from_start);
        p_fifo->write_index = len_from_start;
    }

    p_fifo->free_num -= len;
    p_fifo->used_num += len;
    retval = len;
    CRITICAL_SETCION_EXIT();

    return retval;
}

//******************************************************************************************
//
//! \brief  Put some elements into FIFO, ingnore the interrupt
//!
//! \param  [in]  p_fifo is the pointer of valid FIFO.
//! \param  [in]  p_source is the data element you want to put
//! \param  [in]  the number of elements
//! \retval the number of really write data, otherwise return -1.
//
//******************************************************************************************
int FIFO_S_Puts_NoProtect(FIFO_S_t *p_fifo, uint8_t *p_source, uint32_t len)
{
    int      retval;
    uint32_t len_to_end;
    uint32_t len_from_start;

    ASSERT(NULL != p_fifo);

    if (NULL == p_source)
        return -1;

    if (0 == p_fifo->free_num)
        return 0;

    len = (len < p_fifo->free_num) ? len : p_fifo->free_num;
    len_to_end = p_fifo->buffer_size - p_fifo->write_index;

    if(len_to_end >= len) //no rollback
    {
        len_to_end = len;
        memcpy(&p_fifo->p_start_addr[p_fifo->write_index], p_source, len_to_end);
        p_fifo->write_index = (p_fifo->write_index + len_to_end);
    }
    else  //rollback
    {
        len_from_start = len - len_to_end;
        memcpy(&p_fifo->p_start_addr[p_fifo->write_index], p_source, len_to_end);
        memcpy(&p_fifo->p_start_addr[0], p_source + len_to_end, len_from_start);
        p_fifo->write_index = len_from_start;
    }

    p_fifo->free_num -= len;
    p_fifo->used_num += len;
    retval = len;

    return retval;
}

//******************************************************************************************
//
//! \brief  Get an element from FIFO(in single mode).
//!
//! \param  [in]  p_fifo is the pointer of valid FIFO.
//!
//! \retval the data element of FIFO.
//
//******************************************************************************************
uint8_t FIFO_S_Get(FIFO_S_t* p_fifo)
{
    uint8_t retval = 0;

    //! Check input parameters.
    ASSERT(NULL != p_fifo);

    CRITICAL_SETCION_ENTER();
    retval = p_fifo->p_start_addr[p_fifo->read_index];
    // Update information
    p_fifo->read_index++;
    if(p_fifo->read_index == p_fifo->buffer_size) p_fifo->read_index = 0;
    p_fifo->free_num++;
    p_fifo->used_num--;
    CRITICAL_SETCION_EXIT();

    return (retval);
}

//******************************************************************************************
//
//! \brief  Get some element from FIFO(in single mode).
//!
//! \param  [in]  p_fifo is the pointer of valid FIFO.
//!
//! \retval the number of really read data.
//
//******************************************************************************************
int32_t  FIFO_S_Gets(FIFO_S_t* p_fifo, uint8_t* p_dest, uint32_t len)
{
    uint32_t retval;
    uint32_t len_to_end;
    uint32_t len_from_start;

    ASSERT(NULL != p_fifo);

    if (NULL == p_dest)
        return -1;

    if (0 == p_fifo->used_num)
        return 0;

    CRITICAL_SETCION_ENTER();
    len = (len < p_fifo->used_num) ? len : p_fifo->used_num;
    len_to_end = p_fifo->buffer_size - p_fifo->read_index;

    if(len_to_end >= len) //no rollback
    {
        len_to_end = len;
        memcpy(p_dest, &p_fifo->p_start_addr[p_fifo->read_index], len_to_end);
        p_fifo->read_index = (p_fifo->read_index + len_to_end);
    }
    else  //rollback
    {
        len_from_start = len - len_to_end;
        memcpy(p_dest, &p_fifo->p_start_addr[p_fifo->read_index], len_to_end);
        memcpy(p_dest + len_to_end, &p_fifo->p_start_addr[0], len_from_start);
        p_fifo->read_index = len_from_start;
    }

    p_fifo->free_num += len;
    p_fifo->used_num -= len;
    retval = len;
    CRITICAL_SETCION_EXIT();

    return retval;
}



//******************************************************************************************
//
//! \brief  Pre-Read an element from FIFO(in single mode).
//!
//! \param  [in]  p_fifo is the pointer of valid FIFO.
//! \param  [in]  Offset is the offset from current pointer.
//!
//! \retval the data element of FIFO.
//
//******************************************************************************************
uint8_t FIFO_S_PreRead(FIFO_S_t* p_fifo, uint32_t offset)
{
    uint32_t index;
    uint8_t retval;

    //! Check input parameters.
    ASSERT(NULL != p_fifo);

    if (offset > p_fifo->used_num)
    {
        return 0x00;
    }
    else
    {
        CRITICAL_SETCION_ENTER();
        index = p_fifo->read_index + offset;
        if(index == p_fifo->buffer_size) index = 0;
        // Move Read Pointer to right position
        retval = p_fifo->p_start_addr[index];
        CRITICAL_SETCION_EXIT();

        return retval;
    }
}

/*
 *
 *
 *
 *
 */
int32_t FIFO_S_PreReads(FIFO_S_t* p_fifo, uint8_t *p_dest, uint32_t offset, uint32_t len)
{
    uint32_t retval;
    uint32_t temp_index;
    uint32_t len_to_end;
    uint32_t len_from_start;

    ASSERT(NULL != p_fifo);

    if (NULL == p_dest)
        return -1;

    if (0 == p_fifo->used_num)
        return 0;

    if (offset >= p_fifo->used_num)
        return 0;

    CRITICAL_SETCION_ENTER();
    temp_index = p_fifo->read_index + offset;
    if(temp_index == p_fifo->buffer_size) temp_index = 0;
    len = (len < (p_fifo->used_num - offset)) ? len : (p_fifo->used_num - offset);
    len_to_end = p_fifo->buffer_size - temp_index;

    if(len_to_end >= len) //no rollback
    {
        len_to_end = len;
        memcpy(p_dest, &p_fifo->p_start_addr[temp_index], len_to_end);
    }
    else  //rollback
    {
        len_from_start = len - len_to_end;
        memcpy(p_dest, &p_fifo->p_start_addr[temp_index], len_to_end);
        memcpy(p_dest + len_to_end, &p_fifo->p_start_addr[0], len_from_start);
    }

    retval = len;
    CRITICAL_SETCION_EXIT();

    return retval;
}


//******************************************************************************************
//
//! \brief  FIFO is empty (in single mode)?
//!
//! \param  [in] p_fifo is the pointer of valid FIFO.
//!
//! \retval - None-zero(true) if empty.
//!         - Zero(false) if not empty.
//
//******************************************************************************************
uint8_t FIFO_S_IsEmpty(FIFO_S_t* p_fifo)
{
    uint8_t retval;
    //! Check input parameter.
    ASSERT(NULL != p_fifo);

    //CRITICAL_SETCION_ENTER();
    retval = (p_fifo->used_num ? 0 : 1);
    //CRITICAL_SETCION_EXIT();

    return retval;
}



//******************************************************************************************
//
//! \brief  FIFO is full (in single mode)?
//!
//! \param  [in] p_fifo is the pointer of valid FIFO.
//!
//! \retval - None-zero(true) if full.
//!         - Zero(false) if not full.
//
//******************************************************************************************
uint8_t FIFO_S_IsFull(FIFO_S_t* p_fifo)
{
    uint8_t retval;
    //! Check input parameter.
    ASSERT(NULL != p_fifo);

    //CRITICAL_SETCION_ENTER();
    retval = (p_fifo->free_num ? 0 : 1);
    //CRITICAL_SETCION_EXIT();

    return retval;
}



//******************************************************************************************
//
//! \brief  Get FIFO the number of elements(in single mode)?
//!
//! \param  [in] p_fifo is the pointer of valid FIFO.
//!
//! \retval The number of elements in FIFO.
//
//******************************************************************************************
uint32_t FIFO_S_CountUsed(FIFO_S_t* p_fifo)
{
    uint32_t retval;
    //! Check input parameter.
    ASSERT(NULL != p_fifo);

    //CRITICAL_SETCION_ENTER();
    retval = p_fifo->used_num;
    //CRITICAL_SETCION_EXIT();

    return retval;
}



//******************************************************************************************
//
//! \brief  Get FIFO the number of elements(in single mode)?
//!
//! \param  [in] p_fifo is the pointer of valid FIFO.
//!
//! \retval The number of elements in FIFO.
//
//******************************************************************************************
uint32_t FIFO_S_CountFree(FIFO_S_t* p_fifo)
{
    uint32_t retval;
    //! Check input parameter.
    ASSERT(NULL != p_fifo);


    //CRITICAL_SETCION_ENTER();
    retval = p_fifo->free_num;
    //CRITICAL_SETCION_EXIT();

    return retval;
}


//******************************************************************************************
//
//! \brief  Flush the content of FIFO.
//!
//! \param  [in] p_fifo is the pointer of valid FIFO.
//!
//! \retval 0 if success, -1 if failure.
//
//******************************************************************************************
void FIFO_S_Flush(FIFO_S_t* p_fifo)
{
    //! Check input parameters.
    ASSERT(NULL != p_fifo);
    //! Initialize FIFO Control Block.
    CRITICAL_SETCION_ENTER();
    p_fifo->free_num    = p_fifo->buffer_size;
    p_fifo->used_num    = 0;
    p_fifo->read_index  = 0;
    p_fifo->write_index = 0;
    CRITICAL_SETCION_EXIT();
}

uint32_t FIFO_S_Discard(FIFO_S_t *p_fifo, uint32_t len)
{
    //! Check input parameters.
    uint32_t index;
    ASSERT(NULL != p_fifo);

    CRITICAL_SETCION_ENTER();
    if (len > p_fifo->used_num)
        len = p_fifo->used_num;
    //! Initialize FIFO Control Block.
    index = len + p_fifo->read_index;
    if(index == p_fifo->buffer_size) index = 0;
    p_fifo->read_index = index;
    p_fifo->free_num += len;
    p_fifo->used_num -= len;
    CRITICAL_SETCION_EXIT();
    return len;
}


