//******************************************************************************************
//!
//! \file   FIFO.h
//! \brief  Genernal FIFO Model Interface.
//!         You can use uniform FIFO Model to manager Any type of data element.
//! \author cedar
//! \changed damom.li
//! \changed sky.huang
//! \date   2016-11-03
//! \email  xuesong5825718@gmail.com
//!
//! \license
//! Copyright (c) 2013 Cedar MIT License
//!
//! Permission is hereby granted, free of charge, to any person obtaining a copy
//! of this software and associated documentation files (the "Software"), to deal
//! in the Software without restriction, including without limitation the rights to
//! use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of
//! the Software, and to permit persons to whom the Software is furnished to do so,
//! subject to the following conditions:
//!
//! The above copyright notice and this permission notice shall be included in all
//! copies or substantial portions of the Software.
//!
//! THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
//! IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
//! FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
//! AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
//! LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
//! OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS
//! IN THE SOFTWARE.
///
//******************************************************************************************
#ifndef __FIFO_H__
#define __FIFO_H__
#ifdef __cplusplus
extern "C"
{
#endif

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include "stm32f4xx.h"

//******************************************************************************************
//!                           CONFIGURE MACRO
//******************************************************************************************

//#define USE_DYNAMIC_MEMORY                 //!< Use system malloc/free function
// #define NDEBUG


typedef unsigned char uint8_t;
typedef unsigned int  uint32_t;

#ifdef USE_DYNAMIC_MEMORY
extern void *pvPortMalloc( size_t xSize );
extern void vPortFree( void *pv );
#endif


//******************************************************************************************
//!                           PUBLIC TYPE
//******************************************************************************************

//! FIFO Memory Model

//! FIFO Memory Model (Single Byte Mode)
typedef struct
{
    uint8_t* p_start_addr;                  //!< FIFO Memory Pool Start Address
    uint8_t* p_end_addr;                    //!< FIFO Memory Pool End Address
    uint32_t free_num;                      //!< The remain capacity of FIFO
    uint32_t buffer_size;                   //!< Elements capacity
    uint32_t used_num;                          //!< The number of elements in FIFO
    uint32_t read_index;                    //!< FIFO Data Read Index Pointer
    uint32_t write_index;                   //!< FIFO Data Write Index Pointer
} FIFO_S_t;

//******************************************************************************************
//!                           PUBLIC API
//******************************************************************************************

#ifdef USE_DYNAMIC_MEMORY


//******************************************************************************************
//
//! \brief  Create An New FIFO Instance(in Single Mode).
//! This function allocate enought room for N blocks fifo elements, then return the pointer
//! of FIFO.
//!
//! \param  [in] uint_cnt is count of fifo elements.
//! \retval The Pointer of FIFO instance, return NULL is failure to allocate memory.
//!
//! \note   -# You must enable USE_MEMORY_ALLOC macro and ensure your system have <stdlib.h>
//!            Header file before use this function.
//! \note   -# Functions FIFO_Create and FIFO_Destory must be used in pairs.
//!
//******************************************************************************************
extern FIFO_S_t* FIFO_S_Create(uint32_t uint_cnt);



//******************************************************************************************
//
//! \brief  Destory FIFO Instance(in Single Mode).
//!  This function release memory, then reinit the FIFO struct.
//!
//! \param  [in] p_fifo is the pointer of FIFO instance
//! \retval None.
//!
//! \note   -# You must enable USE_MEMORY_ALLOC macro and ensure your system have <stdlib.h>
//!            Header file before use this function.
//
//******************************************************************************************
extern void FIFO_S_Destory(FIFO_S_t* p_fifo);

#endif // USE_DYNAMIC_MEMORY



//******************************************************************************************
//
//! \brief  Initialize an static FIFO struct(in single mode).
//!
//! \param  [in] p_fifo is the pointer of valid FIFO instance.
//! \param  [in] p_base_addr is the base address of pre-allocate memory, such as array.
//! \param  [in] uint_cnt is count of fifo elements.
//! \retval 0 if initialize successfully, otherwise return -1.
//
//******************************************************************************************
extern int FIFO_S_Init(FIFO_S_t* p_fifo, void* p_base_addr, uint32_t uint_cnt);



//******************************************************************************************
//
//! \brief  Put an element into FIFO(in single mode).
//!
//! \param  [in]  p_fifo is the pointer of valid FIFO.
//! \param  [in]  element is the data element you want to put
//!
//! \retval 0 if operate successfully, otherwise return -1.
//
//******************************************************************************************
extern int FIFO_S_Put(FIFO_S_t* p_fifo, uint8_t element);

extern int FIFO_S_Puts(FIFO_S_t *p_fifo, uint8_t *p_source, uint32_t len);
extern int FIFO_S_Puts_NoProtect(FIFO_S_t *p_fifo, uint8_t *p_source, uint32_t len);

//******************************************************************************************
//
//! \brief  Get an element from FIFO(in single mode).
//!
//! \param  [in]  p_fifo is the pointer of valid FIFO.
//!
//! \retval the data element of FIFO.
//
//******************************************************************************************
extern uint8_t  FIFO_S_Get(FIFO_S_t* p_fifo);
extern int32_t  FIFO_S_Gets(FIFO_S_t* p_fifo, uint8_t* p_dest, uint32_t len);


//******************************************************************************************
//
//! \brief  Pre-Read an element from FIFO(in single mode).
//!
//! \param  [in]  p_fifo is the pointer of valid FIFO.
//! \param  [in]  Offset is the offset from current pointer.
//!
//! \retval the data element of FIFO.
//
//******************************************************************************************
extern uint8_t FIFO_S_PreRead(FIFO_S_t* p_fifo, uint32_t offset);
extern int32_t FIFO_S_PreReads(FIFO_S_t* p_fifo, uint8_t *p_dest, uint32_t offset, uint32_t len);



//******************************************************************************************
//
//! \brief  FIFO is empty (in single mode)?
//!
//! \param  [in] p_fifo is the pointer of valid FIFO.
//!
//! \retval - None-zero(true) if empty.
//!         - Zero(false) if not empty.
//
//******************************************************************************************
extern uint8_t FIFO_S_IsEmpty(FIFO_S_t* p_fifo);



//******************************************************************************************
//
//! \brief  FIFO is full (in single mode)?
//!
//! \param  [in] p_fifo is the pointer of valid FIFO.
//!
//! \retval - None-zero(true) if full.
//!         - Zero(false) if not full.
//
//******************************************************************************************
extern uint8_t FIFO_S_IsFull(FIFO_S_t* p_fifo);



//******************************************************************************************
//
//! \brief  Get FIFO the number of elements(in single mode)?
//!
//! \param  [in] p_fifo is the pointer of valid FIFO.
//!
//! \retval The number of elements in FIFO.
//
//******************************************************************************************
extern uint32_t FIFO_S_CountUsed(FIFO_S_t* p_fifo);



//******************************************************************************************
//
//! \brief  Get FIFO the number of elements(in single mode)?
//!
//! \param  [in] p_fifo is the pointer of valid FIFO.
//!
//! \retval The number of elements in FIFO.
//
//******************************************************************************************
extern uint32_t FIFO_S_CountFree(FIFO_S_t* p_fifo);



//******************************************************************************************
//
//! \brief  Flush the content of FIFO.
//!
//! \param  [in] p_fifo is the pointer of valid FIFO.
//!
//! \retval 0 if success, -1 if failure.
//
//******************************************************************************************
extern void      FIFO_S_Flush(FIFO_S_t* p_fifo);
extern uint32_t  FIFO_S_Discard(FIFO_S_t *p_fifo, uint32_t len);
#ifdef __cplusplus
}
#endif

#endif // __FIFO_H__

